/*
 * chara.c
 *
 * Last modified: <2014/12/30 23:31:22 +0900> By @MIRROR_
 */
#include "nes.h"
#include "game.h"

#ifdef USE_SOUND
#include "nsd.h"
#include "sound.h"
#endif

#include "math.h"
#include "chara.h"
#include "map.h"
#include "palette.h"

#define CHR_CONFUSION 0x3e
#define CHR_FACE      0x46

#define CHT_KICK      10

// LN^e[u
const char charatable[][5] = {
	{0x00, 0x0b,0x0c,0x29,0x2a}, // w
	{0x00, 0x09,0x0a,0x27,0x28}, // wʑOi1
	{0x40, 0x0a,0x09,0x28,0x27}, // wʑOi2

	{0x40, 0x06,0x05,0x24,0x23}, // 
	{0x40, 0x08,0x07,0x26,0x25}, // Oi

	{0x00, 0x01,0x02,0x1f,0x20}, // 
	{0x00, 0x03,0x04,0x21,0x22}, // ʑOi1
	{0x40, 0x04,0x03,0x22,0x21}, // ʑOi2

	{0x00, 0x05,0x06,0x23,0x24}, // E
	{0x00, 0x07,0x08,0x25,0x26}, // EOi

	{0x00, 0x0b,0x0c,0x44,0x45}, // LbN/NA|[Y w
	{0x40, 0x06,0x05,0x43,0x42}, // LbN/NA|[Y 
	{0x00, 0x01,0x02,0x40,0x41}, // LbN/NA|[Y 
	{0x00, 0x05,0x06,0x42,0x43}, // LbN/NA|[Y E
};

// Aj[Ve[u
const char animetable[][4] = {
	{0, 1, 0, 2}, // 
	{3, 4, 3, 4}, // 
	{5, 6, 5, 7}, // 
	{8, 9, 8 ,9}, // E

	{10, 0, 10, 0},
	{11, 3, 11, 3},
	{12, 5, 12, 5},
	{13, 8, 13, 8},
};

// XCLN^e[u
const char slime_chrtbl[][4] = {
	{0x0d,0x0e,0x2b,0x2c}, // 
	{0x0f,0x10,0x2d,0x2e}, // 
	{0x11,0x12,0x2f,0x30}, // 
};

// XCAj[Ve[u
const char slime_anmtbl[][4] = {
	{1, 0, 1, 2}, // 
	{1, 2, 1, 0}, // 
	{1, 0, 1, 2}, // 
	{1, 2, 1, 0}, // E
};

// @gLN^e[u
const char wizard_chrtbl[][5] = {
	{0x00, 0x1b,0x1c,0x39,0x3a}, // w
	{0x00, 0x1d,0x1e,0x3b,0x3c}, // wʑOi1
	{0x40, 0x1e,0x1d,0x3c,0x3b}, // wʑOi2

	{0x40, 0x18,0x17,0x36,0x35}, // E
	{0x40, 0x1a,0x19,0x38,0x37}, // EOi

	{0x00, 0x13,0x14,0x31,0x32}, // 
	{0x00, 0x15,0x16,0x33,0x34}, // ʑOi1
	{0x40, 0x16,0x15,0x34,0x33}, // ʑOi2

	{0x00, 0x17,0x18,0x35,0x36}, // E
	{0x00, 0x19,0x1a,0x37,0x38}, // EOi
};

// LN^pbgwe[u
const unsigned char charapal[] = {
	0x00, // CHR_BLUE_PENGUIN
	0x01, // CHR_RED_PENGUIN
	0x02, // CHR_SLIME
	0x03, // CHR_WIZARD
	0x01, // CHR_RED_SLIME
	0x02, // CHR_SORCERER
	0x01, // CHR_FIRE_BALL
	0x02, // CHR_REVERSE_BALL
};

// OiΈړl
const char forward_offset[][2] = {
	{ 0, -1}, // 
	{ 1,  0}, // E
	{ 0,  1}, // 
	{-1,  0}, // 
};

Chara chara[CHARA_MAX];
Ball ball[BALL_MAX];

void __fastcall__ chara_init(void) {

	unsigned char i;

	// pbgݒ
	SetPalette(palettesp, 1);

	// LN^̏
	FillMemory(chara, 0, sizeof(chara));

	for (i = 0; i < CHARA_MAX; ++i) {
		Chara* p = &chara[i];
		p->chr = CHR_NOTHING;
		p->move.dir.cur = 2;
	}

	// {[̏
	FillMemory(&ball, 0, sizeof(ball));
}

void __fastcall__ chara_draw(unsigned char frame) {
	unsigned char i;

	for (i = 0; i < CHARA_MAX; ++i) {
		Chara *p = &chara[i];

		switch (p->chr) {
		case CHR_BLUE_PENGUIN:
		case CHR_RED_PENGUIN:
			{
				unsigned char x = p->x  + ((BG_FIELD_W-MAP_FIELD_W)/2)*16 + 8;
				unsigned char y = p->y  + ((BG_FIELD_H-MAP_FIELD_H)/2)*16 + 6;
				char id = p->move.dir.cur;
				const char* chara;
				unsigned char flag;

				if (p->status & STATUS_KICK) {
					chara = charatable[CHT_KICK + id];

				} else {
					if (game.task == GAME_TASK_GOAL) {
						id += 4;
						y -= (~frame & 1) << 2;
					}

					id = animetable[id][frame];
					chara = charatable[id];
				}

				flag = chara[0] | charapal[p->chr];

				SetDMA(x-8, y-8, chara[1], flag);
				SetDMA(x  , y-8, chara[2], flag);
				SetDMA(x-8, y  , chara[3], flag);
				SetDMA(x  , y  , chara[4], flag);

				if (game.status & GAME_STATUS_EXCHANGE) {
					unsigned char i = game.frame & 1;
					if (i == p->chr) {
						i = ~i & 1;
						SetDMA(x-4, y-12, CHR_FACE, charapal[i]);
					}
				}
				if (p->status & STATUS_CONFUSION) {
					unsigned char id = CHR_CONFUSION + (frame & 1);
					y -= (game.status & GAME_STATUS_EXCHANGE)? 20: 16;
					SetDMA(x-4, y, id, 0);
				}
				break;
			}

		case CHR_WIZARD:
		case CHR_SORCERER:
			{
				const char  anime = animetable[p->move.dir.cur][frame];
				const char* chara = wizard_chrtbl[anime];
				unsigned char flag = chara[0] | charapal[p->chr];
				unsigned char x = p->x  + ((BG_FIELD_W-MAP_FIELD_W)/2)*16 + 8;
				unsigned char y = p->y  + ((BG_FIELD_H-MAP_FIELD_H)/2)*16 + 6;

				SetDMA(x-8, y-8, chara[1], flag);
				SetDMA(x  , y-8, chara[2], flag);
				SetDMA(x-8, y  , chara[3], flag);
				SetDMA(x  , y  , chara[4], flag);
				break;
			}

		case CHR_SLIME:
		case CHR_RED_SLIME:
			{
				const char  anime = slime_anmtbl[p->move.dir.cur][frame];
				const char *chara = slime_chrtbl[anime];
				unsigned char flag = charapal[p->chr];
				unsigned char x = p->x  + ((BG_FIELD_W-MAP_FIELD_W)/2)*16 + 8;
				unsigned char y = p->y  + ((BG_FIELD_H-MAP_FIELD_H)/2)*16 + 6;

				SetDMA(x-8, y-8, chara[0], flag);
				SetDMA(x  , y-8, chara[1], flag);
				SetDMA(x-8, y  , chara[2], flag);
				SetDMA(x  , y  , chara[3], flag);
				break;
			}

		case CHR_FIRE_BALL:
		case CHR_REVERSE_BALL:
			{
				unsigned char x = p->x  + ((BG_FIELD_W-MAP_FIELD_W)/2)*16 + 4;
				unsigned char y = p->y  + ((BG_FIELD_H-MAP_FIELD_H)/2)*16 + 4;
				unsigned char flag = charapal[p->chr];
				SetDMA(x, y, 0x3d, flag);
			}
			break;
		}
	}
}

void __fastcall__ chara_action_move(Chara* p) {

	// Ԃѐ
	if (p->status & STATUS_LEAP) {
		const char* fo = forward_offset[p->move.dir.ext];

		if (p->status & STATUS_RESET) {
			unsigned char x0, y0, x1, y1;
			unsigned char mcp1;

			// LN^̈ړmF
			x0 = p->x >> 4; x1 = x0 + fo[0];
			y0 = p->y >> 4; y1 = y0 + fo[1];
			map_getmcp(x1, y1, mcp1);

			// Ԃьp
			switch (mcp1) {
			case MCP_FIELD:
			case MCP_GOAL: {
				unsigned char mcp0;

				map_getmcp(x0, y0, mcp0);
				mcp0 = REMOVE_CHAR(mcp0);
				map_setmcp(x0, y0, mcp0);
				mcp1 = MAP_SYNBOL(p->id, GET_CHIP_ID(mcp1));
				map_setmcp(x1, y1, mcp1);
				p->status &= ~STATUS_RESET;
				break;
			}
			default: {
				unsigned char cid = GET_CHAR_ID(mcp1);

				if (cid < CHARA_ID_MAX) {
					Chara* ch = &chara[cid];
					// Ԃѐ悪LN^̏ꍇA]
					chara_action_leap(ch, p->move.dir.ext, p->status & STATUS_LOOP);
				}
				p->status &= ~(STATUS_BUSY | STATUS_LEAP | STATUS_LOOP);
				return;
			}}
		}

		p->x = (p->x & ~0x07) + (fo[0] << 3);
		p->y = (p->y & ~0x07) + (fo[1] << 3);

		if ((p->x & 0x0f) == 0 && (p->y & 0x0f) == 0) {
			if (p->status & STATUS_LOOP) {
				p->status |= STATUS_RESET;
			} else {
				p->status &= ~(STATUS_BUSY | STATUS_LEAP);
			}
		}

	} else {
		// ^]
		const char* fo = forward_offset[p->move.dir.cur];

		p->x += fo[0];
		p->y += fo[1];

		if ((p->x & 0x0f) == 0 && (p->y & 0x0f) == 0) p->status &= ~STATUS_BUSY;
	}
}

char __fastcall__ chara_action_leap(Chara* p, unsigned char dir, unsigned char status) {
	if (p->status & STATUS_LEAP) {
		return 0;
	}
	if (p->status & STATUS_BUSY) {
		const char* fo = forward_offset[p->move.dir.cur];
		char dx = p->x & 0x0f;
		char dy = p->y & 0x0f;

		if (fo[0] < 0) dx |= 0xf0;
		if (fo[1] < 0) dy |= 0xf0;

		if (abs(dx) < 4 || abs(dy) < 4) {
			return 0;
		}

		p->x = (p->x & 0xf0) + fo[0] << 4;
		p->y = (p->y & 0xf0) + fo[1] << 4;
	}

	p->status |= STATUS_LEAP | STATUS_RESET | STATUS_BUSY | status;
	p->move.dir.ext = dir;
	return 1;
}

void __fastcall__ action_penguin(Chara* p) {
	unsigned char i = p->id;
	unsigned char dir = -1;
	unsigned char pad;

	if (game.status & GAME_STATUS_EXCHANGE) {
		pad = (i+1) & 1;
	} else {
		pad = i;
	}

	// L[͂ɂړ
	if      (ButtonPress(pad, BTN_RIGHT)) { dir = 1; }
	else if (ButtonPress(pad, BTN_LEFT )) { dir = 3; }
	else if (ButtonPress(pad, BTN_UP   )) { dir = 0; }
	else if (ButtonPress(pad, BTN_DOWN )) { dir = 2; }

	if (ButtonPress(pad, BTN_A))
		p->status |= STATUS_KICK;
	else
		p->status &= ~STATUS_KICK;


	// ړ̊mF
	if (dir < 4) {

		// ͏㉺Eւ
		if (p->status & STATUS_CONFUSION) {
			dir = (dir + 2) & 3;
		}

		// قȂꍇ́Aς
		if (p->move.dir.cur != dir) {
			/*        dir
			 *        +--+--+--+--+
			 * cur    | 0| 1| 2| 3|
			 *   +================+
			 *   | 0  | -| 1| 1| 3| <= Jڐ
			 *   +    +--+--+--+--+
			 *   |   d| -| 1| 2| 3| <= dir - cur
			 *   +    +--+--+--+--+
			 *   | d+3| -| 4| 5| 6| <= dir - cur + 3
			 *   +    +--+--+--+--+
			 *   |   v| -|+1|+1|-1| <= Zl
			 *   +================+
			 *   | 1  | 0| -| 2| 2|
			 *   +    +--+--+--+--+
			 *   |   d|-1| -| 1| 2|
			 *   +    +--+--+--+--+
			 *   | d+3| 2| -| 4| 5|
			 *   +    +--+--+--+--+
			 *   |   v|-1| -|+1|+1|
			 *   +================+
			 *   | 2  | 3| 1| -| 3|
			 *   +    +--+--+--+--+
			 *   |   d|-2|-1| -| 1|
			 *   +    +--+--+--+--+
			 *   | d+3| 1| 2| -| 4|
			 *   +    +--+--+--+--+
			 *   |   v|+1|-1| -|+1|
			 *   +================+
			 *   | 3  | 0| 0| 2| -|
			 *   +    +--+--+--+--+
			 *   |   d|-3|-2|-1| -|
			 *   +    +--+--+--+--+
			 *   | d+3| 0| 1| 2| -|
			 *   +    +--+--+--+--+
			 *   |   v|+1|+1|-1| -|
			 *   +----+--+--+--+--+
			 *
			 *    0  1  2  3  4  5  6  7 <= dir - cur + 3
			 *   +1 +1 -1    +1 +1 -1    <= Zl
			 */
			dir = (dir - p->move.dir.cur + 3) & 3;
			if (dir < 2) dir =  1;
			else         dir = -1;
			p->move.dir.cur = (p->move.dir.cur + dir) & 3;

		} else {
			// LN^ʒũ}bv擾
			unsigned char x0, y0, x1, y1;
			const char* fo = forward_offset[dir];
			unsigned char mcp1;

			x0 = p->x >> 4; x1 = x0 + fo[0];
			y0 = p->y >> 4; y1 = y0 + fo[1];
			map_getmcp(x1, y1, mcp1);

			if (mcp1 == MCP_FIELD || mcp1 == MCP_GOAL) {
				unsigned char mcp0;

				map_getmcp(x0, y0, mcp0);
				mcp0 = REMOVE_CHAR(mcp0);
				map_setmcp(x0, y0, mcp0);
				mcp1 = MAP_SYNBOL(i, GET_CHIP_ID(mcp1));
				map_setmcp(x1, y1, mcp1);

				p->status |= STATUS_BUSY;
			}
		}
	}
	else if (ButtonDown(pad, BTN_A)) {
		// is̃}bv擾
		const char* fo = forward_offset[p->move.dir.cur];
#if defined(BALL_MAX) && (BALL_MAX < 2)
		Ball*b = &ball[0];
#else
		Ball*b = &ball[i];
#endif
		unsigned char x1, y1;
		unsigned char mcp1;

		x1 = p->x >> 4;
		y1 = p->y >> 4;

		// 2P̏ꍇA^̃S[ړ\
		if (i && !(b->status & STATUS_BUSY)) {
			map_getmcp(x1, y1, mcp1);
			if (REMOVE_CHAR(mcp1) == MCP_GOAL) {
				unsigned char x2, y2;
				unsigned char mcp2;

				// ړmF
				x2 = x1 + fo[0];
				y2 = y1 + fo[1];
				map_getmcp(x2, y2, mcp2);

				if (mcp2 == MCP_FIELD) {
#ifdef USE_SOUND
					nsd_play_se(SE_CLICK);
#endif
					b->status = STATUS_BUSY;
					b->x = x1;
					b->y = y1;
					b->dir = p->move.dir.cur;
					map_move(x1, y1, b->dir);
					return;
				}
			}
		}

		x1 += fo[0];
		y1 += fo[1];
		map_getmcp(x1, y1, mcp1);

		if (i) {
			// 2ṔABLOCKFBLOCK̈t]
			// GOALFBLOCKƓňړ
			switch (mcp1) {
			case MCP_BLOCK:
			case MCP_GOAL:
				mcp1 = MCP_FBLOCK;
				break;

			case MCP_FBLOCK:
				mcp1 = MCP_BLOCK;
				break;
			}
		}

		switch (mcp1) {
		case MCP_BLOCK:
#ifdef USE_SOUND
			nsd_play_se(SE_BREAK);
#endif
			map_set(x1, y1, MCP_FIELD);
			break;

		case MCP_FBLOCK:
			if (!(b->status & STATUS_BUSY)) {
				unsigned char x2, y2;
				unsigned char mcp2;

				b->dir = p->move.dir.cur;
				b->x = x1;
				b->y = y1;

				// ړmF
				x2 = x1 + fo[0];
				y2 = y1 + fo[1];
				map_getmcp(x2, y2, mcp2);

				switch (mcp2) {
				case MCP_FIELD:
#ifdef USE_SOUND
					nsd_play_se(SE_CLICK);
#endif
					b->status = STATUS_BUSY;
					map_move(x1, y1, b->dir);
					break;

				case MCP_BALL:
					b->status = STATUS_RESET | STATUS_BALL | STATUS_LEAP | STATUS_BUSY;
					break;

				case MCP_WALL:
					break;

				default:
					b->status = STATUS_RESET | STATUS_LEAP | STATUS_BUSY;
					break;
				}
			}
			break;

		case MCP_ICECUBE:
			if (!(b->status & STATUS_BUSY)) {
				unsigned char x2, y2;
				unsigned char mcp2;

				b->dir = p->move.dir.cur;
				b->x = x1;
				b->y = y1;

				// ړmF
				x2 = x1 + fo[0];
				y2 = y1 + fo[1];
				map_getmcp(x2, y2, mcp2);

				switch (mcp2) {
				case MCP_FIELD:
#ifdef USE_SOUND
					nsd_play_se(SE_CLICK);
#endif
					b->status = STATUS_LOOP | STATUS_BUSY;
					map_move(x1, y1, b->dir);
					break;

				case MCP_BALL:
					b->status = STATUS_RESET | STATUS_BALL | STATUS_LEAP | STATUS_BUSY;
					break;

				case MCP_WALL:
					break;

				default:
					b->status = STATUS_RESET | STATUS_LEAP | STATUS_BUSY;
					break;
				}
			}
			break;

		case MCP_BALL:
			// 1P݈̂ړ
			if (i == 0 && !(b->status & STATUS_BUSY)) {
				unsigned char x2, y2;
				unsigned char mcp2;

				// {[̐mF
				x2 = x1 + fo[0];
				y2 = y1 + fo[1];
				map_getmcp(x2, y2, mcp2);

				switch (mcp2) {
				case MCP_GOAL:
					b->status = STATUS_BALL | STATUS_GOAL | STATUS_LOOP | STATUS_BUSY;
					b->dir = p->move.dir.cur;
					break;

				case MCP_FIELD:
					b->status = STATUS_BALL | STATUS_LOOP | STATUS_BUSY;
					b->dir = p->move.dir.cur;
					break;

				default:
					{
						// isǂꍇAoEh
						unsigned char dir = (p->move.dir.cur + 2) & 3;

						// ̌mF
						fo = forward_offset[dir];
						x2 = (p->x >> 4) + fo[0];
						y2 = (p->y >> 4) + fo[1];
						map_getmcp(x2, y2, mcp2);

						switch (mcp2) {
						case MCP_FIELD:
						case MCP_GOAL:
							b->dir = dir;
							b->status = STATUS_BALL | STATUS_BUSY;

							x2 -= fo[0];
							y2 -= fo[1];
							map_getmcp(x2, y2, mcp2);
							if (REMOVE_CHAR(mcp2) == MCP_GOAL) {
								b->status |= STATUS_GOAL;
							}

							// {[ƈꏏɃvC[ړ
							chara_action_leap(p, dir, 0);
							chara_action_move(p);
							break;

						default:
							// oEhłȂꍇ͈ړȂ
							return;
						}
					}
				}
#ifdef USE_SOUND
				nsd_play_se(SE_CLICK);
#endif
				b->x = x1;
				b->y = y1;
				map_move(x1, y1, b->dir);
			}
			break;

		default:
			{
				unsigned char cid = GET_CHAR_ID(mcp1);

				if (cid < CHARA_ID_MAX) {
					Chara* ch = &chara[cid];
					unsigned char x2, y2;
					unsigned char mcp2;

					// LN^̐mF
					x2 = x1 + fo[0];
					y2 = y1 + fo[1];
					map_getmcp(x2, y2, mcp2);

					// Ԃёݒ
					switch (mcp2) {
					case MCP_FIELD:
					case MCP_GOAL:
#ifdef USE_SOUND
						nsd_play_se(SE_CLICK);
#endif
						chara_action_leap(ch, p->move.dir.cur, STATUS_LOOP);
						break;

					default:
						{
							// isǂꍇAoEh
							unsigned char dir = (p->move.dir.cur + 2) & 3;

							// ̌mF
							fo = forward_offset[dir];
							x2 = (p->x >> 4) + fo[0];
							y2 = (p->y >> 4) + fo[1];
							map_getmcp(x2, y2, mcp2);

							switch (mcp2) {
							case MCP_FIELD:
							case MCP_GOAL:
								// ɃvC[ړ
								if (chara_action_leap(p, dir, 0)) {
									chara_action_move(p);
									chara_action_leap(ch, dir, 0);
								}
								break;
							}
						}
						break;
					}
				}
			}
			break;
		}
	}
}



enum {
	SLIME_ACTION_INIT =  0,
	SLIME_ACTION_WAIT,
	SLIME_ACTION_TURN,
	SLIME_ACTION_MOVE,
	SLIME_ACTION_FIRE,
	SLIME_ACTION_MAX
};

void __fastcall__  action_slime(Chara* s) {
	unsigned char i;

	switch (s->action) {
	case SLIME_ACTION_INIT:
		s->action = SLIME_ACTION_WAIT;

	case SLIME_ACTION_WAIT:
		i = s->id << 3;

		if (game.frame == i) {
			// ߂̃vC[ɋ߂Â
			Vector v[2];
			unsigned char d0, d1;

			v[0].x = (chara[0].x - s->x) >> 4;
			v[0].y = (chara[0].y - s->y) >> 4;
			d0 = abs(v[0].x) + abs(v[0].y);

			v[1].x = (chara[1].x - s->x) >> 4;
			v[1].y = (chara[1].y - s->y) >> 4;
			d1 = abs(v[1].x) + abs(v[1].y);

			i = (d0 <= d1)? 0: 1;
			normal(&v[i]);

			// vC[ɕ
			if (v[i].x <= -4) {
				s->move.dir.cur = 3;
			} else if (v[i].x >= 4) {
				s->move.dir.cur = 1;
			} else if (v[i].y > 0) {
				s->move.dir.cur = 2;
			} else {
				s->move.dir.cur = 0;
			}

			if (s->chr == CHR_RED_SLIME && !(rand() & 3)) {
				s->action = SLIME_ACTION_FIRE;
			} else {
				s->action = SLIME_ACTION_MOVE;
			}
		}
		break;

	case SLIME_ACTION_MOVE:
		{
			// LN^ʒũ}bv擾
			unsigned char x0, y0, x1, y1;
			const char* fo = forward_offset[s->move.dir.cur];
			unsigned char mcp1;

			x0 = s->x >> 4; x1 = x0 + fo[0];
			y0 = s->y >> 4; y1 = y0 + fo[1];
			map_getmcp(x1, y1, mcp1);

			if (mcp1 == MCP_FIELD || mcp1 == MCP_GOAL) {
				unsigned char mcp0;

				map_getmcp(x0, y0, mcp0);
				mcp0 = REMOVE_CHAR(mcp0);
				map_setmcp(x0, y0, mcp0);
				mcp1 = MAP_SYNBOL(s->id, GET_CHIP_ID(mcp1));
				map_setmcp(x1, y1, mcp1);

				s->status |= STATUS_BUSY;
				s->action = SLIME_ACTION_WAIT;

			} else {
				s->action = SLIME_ACTION_TURN;
			}
		}
		break;

	case SLIME_ACTION_TURN:
		i = s->id << 3;

		if (game.frame == i) {
			s->move.dir.cur = (s->move.dir.cur + 1) & 3;
			s->action = SLIME_ACTION_MOVE;
		}
		break;

	case SLIME_ACTION_FIRE:
		// t@C[{[I
		{
			const char* fo = forward_offset[s->move.dir.cur];
			Chara* fireball;

			s->action = SLIME_ACTION_WAIT;

			// Ǘ̈͂QȂ̂ŗ͋Z
			if (chara[FIRE_BALL_TOP].chr == CHR_NOTHING) {
				fireball = &chara[FIRE_BALL_TOP];
			} else if (chara[FIRE_BALL_TOP+1].chr == CHR_NOTHING) {
				fireball = &chara[FIRE_BALL_TOP+1];
			} else {
				break;
			}

			// ˏo߂
			fireball->move.vec.x = fo[0] << 3;
			fireball->move.vec.y = fo[1] << 3;

			// ˏo
			fireball->chr = CHR_FIRE_BALL;
			fireball->status = 0;
			fireball->action = 0;
			fireball->x = s->x;
			fireball->y = s->y;
		}
		break;
	}
}



enum {
	WIZARD_ACTION_INIT =  0,
	WIZARD_ACTION_WAIT,
	WIZARD_ACTION_FIRE,
	WIZARD_ACTION_MAX
};

void __fastcall__  action_wizard(Chara* w) {

	switch (w->action) {
	case WIZARD_ACTION_INIT:
		w->action = WIZARD_ACTION_WAIT;

	case WIZARD_ACTION_WAIT:
		{
			Vector vec;

			vec.x = (chara[0].x - w->x) >> 4;
			vec.y = (chara[0].y - w->y) >> 4;
			normal(&vec);

			// vC[ɕ
			if (vec.x < -4) {
				w->move.dir.cur = 3;
			} else if (vec.x > 4) {
				w->move.dir.cur = 1;
			} else if (vec.y > 0) {
				w->move.dir.cur = 2;
			} else {
				w->move.dir.cur = 0;
			}

			// ˃V[PXֈڍs
			if (game.tcnt == 0) w->action = WIZARD_ACTION_FIRE;
		}
		break;

	case WIZARD_ACTION_FIRE:
		// t@C[{[I
		{
			Chara* fireball;

			w->action = WIZARD_ACTION_WAIT;

			// Ǘ̈͂QȂ̂ŗ͋Z
			if (chara[FIRE_BALL_TOP].chr == CHR_NOTHING) {
				fireball = &chara[FIRE_BALL_TOP];
			} else if (chara[FIRE_BALL_TOP+1].chr == CHR_NOTHING) {
				fireball = &chara[FIRE_BALL_TOP+1];
			} else {
				break;
			}

			// ˏo߂
			fireball->move.vec.x = (chara[0].x - w->x) >> 4;
			fireball->move.vec.y = (chara[0].y - w->y) >> 4;
			normal(&fireball->move.vec);
			fireball->move.vec.x >>= 1;
			fireball->move.vec.y >>= 1;

			// ˏo
			fireball->chr = (w->chr == CHR_WIZARD)? CHR_FIRE_BALL: CHR_REVERSE_BALL;
			fireball->status = 0;
			fireball->action = 0;
			fireball->x = w->x;
			fireball->y = w->y;
		}
		break;
	}
}



enum {
	MASICBALL_ACTION_INIT =  0,
	MASICBALL_ACTION_MOVE,
	MASICBALL_ACTION_WAIT,
	MASICBALL_ACTION_STOP,
	MASICBALL_ACTION_MAX
};

void __fastcall__  action_masic_ball(Chara* b) {
	switch (b->action) {
	case MASICBALL_ACTION_INIT:
		b->action = MASICBALL_ACTION_MOVE;

	case MASICBALL_ACTION_MOVE:
		{
			unsigned char x, y;
			unsigned char mcp;
			unsigned char id;

			// 2t[Ɉړ
			b->action = MASICBALL_ACTION_WAIT;

			// ˏoێ܂܈ړI
			b->x += b->move.vec.x; x = b->x >> 4;
			b->y += b->move.vec.y; y = b->y >> 4;
			map_getmcp(x, y, mcp);
			id = GET_CHIP_ID(mcp);

			if (id != MCP_FIELD) {

#ifdef USE_SOUND
				nsd_play_se(SE_BREAK);
#endif
				b->action = MASICBALL_ACTION_STOP;

				// ړ̏ꍇ͓ŏI
#if defined(BALL_MAX) && (BALL_MAX < 2)
				if ((ball[0].status & STATUS_BUSY) && ball[0].x == x && ball[0].y == y) break;
#else
				if (((ball[0].status & STATUS_BUSY) && ball[0].x == x && ball[0].y == y)
					|| ((ball[1].status & STATUS_BUSY) && ball[1].x == x && ball[1].y == y)) break;
#endif
				switch (id) {
				case MCP_FBLOCK:
					// Xn
					map_set(x, y, MCP_BLOCK);
					break;

				case MCP_ICECUBE:
					// Xn
					map_set(x, y, MCP_FBLOCK);
					break;

				case MCP_BLOCK:
					// 
					if (b->chr == CHR_REVERSE_BALL) {
						map_set(x, y, MCP_ICECUBE);
					}
					break;
				}

			} else {
				unsigned char cid = GET_CHAR_ID(mcp);

				if (cid < CHARA_ID_MAX) {
					Chara* ch = &chara[cid];

					if (ch->chr <= CHR_PLAYERS) {
						char dx = ch->x - b->x;
						char dy = ch->y - b->y;

						if (-6 <= dx && dx <= 6 && -6 <= dy && dy <= 6) {
							if (b->chr == CHR_FIRE_BALL) {
								// Ԃ̐ݒA
								if (ch->status & STATUS_CONFUSION)
									ch->status &= ~STATUS_CONFUSION;
								else
									ch->status |= STATUS_CONFUSION;
							} else {
								// pbhύX̐ݒA
								if (game.status & GAME_STATUS_EXCHANGE)
									game.status &= ~GAME_STATUS_EXCHANGE;
								else
									game.status |= GAME_STATUS_EXCHANGE;
							}
#ifdef USE_SOUND
							nsd_play_se(SE_BREAK);
#endif
							b->action = MASICBALL_ACTION_STOP;
						}
					}
				}
			}
		}
		break;

	case MASICBALL_ACTION_WAIT:
		b->action = MASICBALL_ACTION_MOVE;
		break;

	case MASICBALL_ACTION_STOP:
		b->chr = CHR_NOTHING;
		b->action = MASICBALL_ACTION_INIT;
		b->status = 0;
		break;
	}
}

void __fastcall__ action_ball(Ball* b) {
	// {[̈ړ
	if (b->status & STATUS_RESET) {
		// ړ\mF
		const char* fo = forward_offset[b->dir];
		unsigned char x, y;
		unsigned char mcp;

		b->status &= ~STATUS_RESET;
		b->x += fo[0]; x = b->x + fo[0];
		b->y += fo[1]; y = b->y + fo[1];
		map_getmcp(x, y, mcp);

		switch (mcp) {
		case MCP_GOAL:
			if (b->status & STATUS_BALL) {
				b->status |= STATUS_GOAL;
				map_move(b->x, b->y, b->dir);
				return;
			}
			break;

		case MCP_FIELD:
			map_move(b->x, b->y, b->dir);
			return;

		case MCP_WALL:
			break;

		default:
			if (b->status & (STATUS_LOOP|STATUS_LEAP)) {
				unsigned char x2, y2;
				unsigned char mcp2;

				b->x = x;
				b->y = y;

				switch (mcp) {
				case MCP_ICECUBE:
					// ړmF
					x2 = x + fo[0];
					y2 = y + fo[1];
					map_getmcp(x2, y2, mcp2);

					switch (mcp2) {
					case MCP_FIELD:
						b->status &= ~STATUS_BALL;
						map_move(x, y, b->dir);
						return;

					case MCP_BALL:
						b->status |= STATUS_RESET | STATUS_BALL;
#ifdef USE_SOUND
						nsd_play_se(SE_CLICK);
#endif
						return;

					case MCP_ICECUBE:
						b->status |= STATUS_RESET;
						b->status &= ~STATUS_BALL;
#ifdef USE_SOUND
						nsd_play_se(SE_CLICK);
#endif
						return;
					}
					break;

				case MCP_BALL:
					// {[̐mF
					x2 = x + fo[0];
					y2 = y + fo[1];
					map_getmcp(x2, y2, mcp2);

					switch (mcp2) {
					case MCP_GOAL:
						b->status |= STATUS_GOAL;

					case MCP_FIELD:
						b->status |= STATUS_BALL;
						break;

					case MCP_ICECUBE:
						b->status |= STATUS_RESET;
						b->status &= ~STATUS_BALL;
#ifdef USE_SOUND
						nsd_play_se(SE_CLICK);
#endif
						return;

					default:
						b->status = 0;
						return;
					}
					map_move(x, y, b->dir);
					return;

				case MCP_WALL:
					break;

				default:
					{
						unsigned char cid = GET_CHAR_ID(mcp);

						// ړ悪LN^̏ꍇA]
						if (cid < CHARA_ID_MAX) {
							Chara* ch = &chara[cid];
							chara_action_leap(ch, b->dir, b->status & STATUS_LOOP);
#ifdef USE_SOUND
							nsd_play_se(SE_CLICK);
#endif
						}
						break;
					}
				}
			}
		}

	} else if (b->status) {
		map_move(b->x, b->y, b->dir);

		if (b->status & STATUS_GOAL) {
			// S[ݒ
			game.task = GAME_TASK_GOAL;
			game.frame = 0;
			game.time = 0;
			game.tcnt = 0;
#ifdef USE_SOUND
			nsd_play_bgm(BGM_CLEAR);
#endif

		} else if (b->status & STATUS_LOOP) {
			// JԂړݒ
			b->status |= STATUS_RESET;
			return;
		}
	}

	// ړ~
	b->status = 0;
}
