/*
 * game.c
 *
 * Last modified: <2014/12/30 23:39:11 +0900> By @MIRROR_
 */
#include "nes.h"
#include "game.h"

#ifdef USE_SOUND
#include "nsd.h"
#include "sound.h"
#endif

#include "math.h"
#include "chara.h"
#include "map.h"
#include "demo.h"

const unsigned char msg_scene[] = "SCENE";
const unsigned char msg_clear[] = "CLEAR!";
const unsigned char msg_over[]  = "GAME OVER";
const unsigned char msg_pause[] = "PAUSE";
const unsigned char msg_congraturation[] = "CONGRATURATION!";
const unsigned char msg_thanks[] = "Thanks for your playing!";
const unsigned char msg_demo[] = "DEMO";

GamePrm game;

#ifdef USE_PADREC
unsigned char* padrec0;
unsigned char* padrec1;
unsigned char padrec;
unsigned char pad0_bits;
unsigned char pad0_cnt;
unsigned char pad1_bits;
unsigned char pad1_cnt;
#endif

// C
void __fastcall__ GameMain(void)
{
	// ϐ
	game.time   = 99;
	game.ecnt   = CHARA_ENEMY_TOP;
	game.task   = GAME_TASK_START;
	game.status &= ~(GAME_STATUS_PAUSE|GAME_STATUS_EXCHANGE);
	//game.scene  = 0;

#ifdef USE_PADREC
	padrec0 = (unsigned char*)0x7000;
	padrec1 = (unsigned char*)0x7800;
	padrec = 1;
	pad0_bits = 0;
	pad0_cnt  = 0;
	pad1_bits = 0;
	pad1_cnt  = 0;

	{
		unsigned char* p = (unsigned char*)0x7000;

		while (p < (unsigned char*)0x8000) {
			*p++ = 0xff;
		}
	}
#endif

	// PPȔ
	InitPPU();

#ifdef USE_SOUND
	// hCo
	nsd_init();
#endif

	// DMA
	InitDMA();

	// BG]R}ḧ揉
	InitCMD();

	// 
	srand(7);

	// LN^
	chara_init();

	// }bv
	if (game.status & GAME_STATUS_PLAYDEMO) {
		demo_init();

	} else {
		map_init(game.scene);
	}

	// wi̕\ʒuݒ
	SetScroll(0, 0);

	// `Jn(NMI荞ݗL)
	SetPPU(0x88, 0x1e);

	while (1) {
		// ݏI҂
		Sync();

		// BG]R}h
		ResetCMD();

		// DMA]\񏉊
		ResetDMA();

		if (game.status & GAME_STATUS_PLAYDEMO) {
			// vCf
			if (demo_sync()) {
				return;
			}
			if (game.frame & 0x80) sprint(32, SCREEN_H-28, msg_demo);

		} else {
			// Rg[mF
			CheckPad();
		}

		// Q[Ŝ̐
		switch (game.task) {
		default:
		case GAME_TASK_START:
			if (game.frame >= SCENE_START_WAIT) {
				game.task = GAME_TASK_PLAY;

#ifdef USE_SOUND
				// tJn
				nsd_play_bgm(BGM_PENGUIN);
#endif

			} else {
				sprint((SCREEN_W-7*8)/2, SCREEN_H/2, msg_scene);
				sprint_n2((SCREEN_W-7*8)/2+6*8, SCREEN_H/2, game.scene+1);
			}
			break;

		case GAME_TASK_PLAY:
			// ^C\
			sprint_n02((SCREEN_W-2*8)/2, 14, game.time);

			if (game.status & GAME_STATUS_PAUSE) {
				// |[Y
				if (ButtonDown(0, BTN_START)) {
					game.status &= ~GAME_STATUS_PAUSE;
#ifdef USE_SOUND
					// tĊJ
					nsd_resume_bgm();
#endif

				} else {
					sprint((SCREEN_W-5*8)/2, SCREEN_H/2, msg_pause);
				}
			} else {
				// |[Y
				if (ButtonDown(0, BTN_START)) {
					game.status |= GAME_STATUS_PAUSE;
#ifdef USE_SOUND
					// tꎞ~
					nsd_pause_bgm();
#endif
				} else {

#ifdef USE_PADREC
					if (padrec) {
						unsigned char pad;

						pad = padinfo[0][BTN_STAT_PRESS];
						if (pad0_bits == pad) {
							if (++pad0_cnt == 0xff) {
								*padrec0++ = pad0_bits;
								*padrec0++ = pad0_cnt;
								pad0_cnt = 0;
							}
						} else {
							*padrec0++ = pad0_bits;
							*padrec0++ = pad0_cnt;
							pad0_bits = pad;
							pad0_cnt = 1;
						}

						pad = padinfo[1][BTN_STAT_PRESS];
						if (pad1_bits == pad) {
							if (++pad1_cnt == 0xff) {
								*padrec1++ = pad1_bits;
								*padrec1++ = pad1_cnt;
								pad1_cnt = 0;
							}
						} else {
							*padrec1++ = pad1_bits;
							*padrec1++ = pad1_cnt;
							pad1_bits = pad;
							pad1_cnt = 1;
						}
					}
#endif

					if (ButtonDown(0, BTN_B) || ButtonDown(1, BTN_B)) {
						if (game.status & GAME_STATUS_EXCHANGE) {
							game.status &= ~GAME_STATUS_EXCHANGE;
						} else {
							game.status |= GAME_STATUS_EXCHANGE;
						}
					}
				}
			}
			break;

		case GAME_TASK_GOAL:
			if (game.time >= SCENE_GOAL_WAIT) {
				game.task = GAME_TASK_CLEAR;
			} else {
				if (++game.tcnt >= 60) {
					game.tcnt = 0;
					++game.time;
				}
				sprint((SCREEN_W-6*8)/2, SCREEN_H/2, msg_clear);
			}
			break;

		case GAME_TASK_CLEAR:
#ifdef USE_PADREC
			padrec = 0;
			*padrec0++ = pad0_bits;
			*padrec0++ = pad0_cnt;
			*padrec1++ = pad1_bits;
			*padrec1++ = pad1_cnt;
#endif

			// vCfI
			if (game.status & GAME_STATUS_PLAYDEMO) {
				return;
			}

			// V[XV
			++game.scene;

			// PPȔ
			InitPPU();

			// DMA
			InitDMA();

			// LN^
			chara_init();

			// ϐ
			game.status = game.frame = game.tcnt = 0;
			game.time = 99;

			// I[NA
			if (game.scene >= SCENE_MAX) {
				game.task = GAME_TASK_CONGRATULATION;
				game.scene = 0;
				game.level = 1;

				FillNameTbl(0, 0, BG_NAMTBL_W, BG_NAMTBL_H, 0x00);
				SetNameTbl(msg_congraturation, (BG_NAMTBL_W-15)/2, BG_NAMTBL_H/2-4, 15, 1);
				SetNameTbl(msg_thanks, (BG_NAMTBL_W-24)/2, BG_NAMTBL_H/2-2, 24, 1);
				FillAttribute(BG1_ATTR_HADR, BG1_ATTR_LADR, 0, 64);

#ifdef USE_SOUND
				// tJn
				nsd_play_bgm(BGM_PENGUIN);
#endif
				{
					unsigned char i;
					unsigned char dx = 8;
					unsigned char dy = (MAP_FIELD_H << 3) + 16;

					for (i = 0; i < CHR_PC_NPC_MAX; ++i) {
						Chara* p = &chara[i];

						p->chr = i;
						p->y = dy;

						if (i & 1) {
							p->x = ((MAP_FIELD_W << 3) - 16) + dx;
							dx += 8;
							dy -= 8;
						} else {
							p->x = ((MAP_FIELD_W << 3) - 16) - dx;
						}
					}
				}
			} else {
#ifdef USE_SOUND
				// t~
				nsd_stop_se();
				nsd_stop_bgm();
#endif

				game.task   = GAME_TASK_START;

				// }bv
				map_init(game.scene);
			}

			// wi̕\ʒuݒ
			SetScroll(0, 0);

			// `Jn(NMI荞ݗL)
			SetPPU(0x88, 0x1e);

			continue;

		case GAME_TASK_CONGRATULATION:
			if (ButtonDown(0, BTN_START)) {
#ifdef USE_SOUND
				// SEABGM~
				nsd_stop_se();
				nsd_stop_bgm();
#endif
				return;
			}
			break;

		case GAME_TASK_OVER:
			if (game.frame > SCENE_OVER_WAIT) {
#ifdef USE_SOUND
				// SEABGM~
				nsd_stop_se();
				nsd_stop_bgm();
#endif
				return;

			} else {
				sprint((SCREEN_W-9*8)/2, SCREEN_H/2, msg_over);
			}
		}

		// ^C\
		//sprint_n02((SCREEN_W-2*8)/2, 14, game.time);

		// LN^̐
		if (!(game.status & GAME_STATUS_PAUSE)) {

			if (game.task == GAME_TASK_PLAY) {
				unsigned char i;

				// {[̈ړ
				action_ball(&ball[0]);
#if defined(BALL_MAX) && (BALL_MAX > 1)
				action_ball(&ball[1]);
#endif
				// LN^̐
				for (i = 0; i < CHARA_MAX; ++i) {
					Chara* p = &chara[i];

					if (p->status & STATUS_BUSY) {
						// ^]
						chara_action_move(p);

					} else {
						switch (p->chr) {
						case CHR_BLUE_PENGUIN:
						case CHR_RED_PENGUIN:
							action_penguin(p);
							break;

						case CHR_SLIME:
						case CHR_RED_SLIME:
							action_slime(p);
							break;

						case CHR_WIZARD:
						case CHR_SORCERER:
							action_wizard(p);
							break;

						case CHR_FIRE_BALL:
						case CHR_REVERSE_BALL:
							action_masic_ball(p);
							break;
						}
					}
				}

				// Ԃ̌Z
				if (game.time > 0 && ++game.tcnt >= 100) {
					game.tcnt = 0;
					--game.time;

					// ^CI[o[
					if (game.time == 0) {
						game.task = GAME_TASK_OVER;
						game.frame = 0;
					}
				}
			}
			// t[JE^XV
			++game.frame;
		}

		// LN^̕\
		chara_draw((game.frame >> 4) & 3);

		// DMA]f[^ݒI
		SetEndDMA();

		// BG]f[^ݒI
		SetEndCMD();
	}
}
